
(* Modify only typecheck and interpret 
 * Though for fun you should try both environment implementations.
 *)

open Ast

exception TypeError
exception RunTimeError
exception Unimplemented (* shows you what to change *)

(*
(* commented out, but an approach equivalent to the one below using
   explicit mutation *)
(*********** environment implementation where env = var->exp **********)

(* Note: types do not matter at run-time, so we just use Ignore to fill 
   in for a type we do not care about *) 

let mt_env x = raise RunTimeError

let extend_env env x e = fun y -> if x=y then e else env y

(* call this from the Rec case of interpret *)
let extend_env_rec env f x e = 
  let rec new_env y =
    if y=f
    then Closure(Fun(x,Ignore,e), new_env)
    else env y
  in new_env

let lookup_env env y = env y
*)

(******* environment implementation where env = (var * (exp ref)) list *******)

(* Note: types do not matter at run-time, so we just use Ignore to fill 
   in for a type we do not care about *) 

let mt_env = []

let extend_env env x e = (x,ref e)::env

(* call this from the Rec case of interpret *)
let extend_env_rec env f x e =
  let r       = ref True in
  let new_env = (f,r)::env in
  let _       = r := Closure(Fun(x,Ignore,e),new_env) in
  new_env

let rec lookup_env env y =
  match env with
    [] -> raise RunTimeError
  | (x,r)::tl -> if x=y then !r else lookup_env tl y

(* evaluator -- remember we are large-step -- return a value 
   we have done two cases for you *)
let rec interpret env e =
  match e with
    True -> e
  | Plus(e1,e2) ->
      let v1 = interpret env e1 in
      let v2 = interpret env e2 in
      (match (v1,v2) with
	(Const i1, Const i2) -> Const(i1+i2)
      |	_ -> raise RunTimeError)
  | _ -> raise Unimplemented

(* type-checker -- we have done three cases for you *)
(* Note: t1=t2 will do what you want to compare types structurally, 
   e.g., Arrow(Int,Bool)=Arrow(Int,Bool)
   But t1==t2 is roughly pointer-equality -- do not accidentally use it or
   you will be very sorry *)

let mt_ctxt                = raise Unimplemented
let extend_ctxt ctxt x t   = raise Unimplemented
let rec lookup_ctxt ctxt y = raise Unimplemented

let rec typecheck ctxt e =
  match e with
    True  -> Bool
  | Plus(e1,e2) ->
      let t1 = typecheck ctxt e1 in
      let t2 = typecheck ctxt e2 in
      if (t1=Int) && (t2=Int)
      then Int
      else raise TypeError
  | Closure(_,_) -> raise TypeError
  | _ -> raise Unimplemented

(** Do not change anything below here **)

let parse_file () =
  let argv = Sys.argv in
  let _ = 
    if Array.length argv != 2
    then (prerr_string ("usage: " ^ argv.(0) ^ " [file-to-parse]\n");
	  exit 1) in
  let ch = open_in argv.(1) in
  Parse.program Lex.lexer (Lexing.from_channel ch)

let _ =
  let prog = parse_file() in
  (try
    (ignore(typecheck mt_ctxt prog);
(*     print_string (exp2string prog); *) (* useful for debugging *)
     print_string ("\ninterpreted answer = ");
     print_string (exp2string (interpret mt_env prog));
     print_newline ())
  with TypeError -> print_string "did not typecheck\n")
